@@ -4,7 +4,7 @@ use strict;
 use warnings FATAL => 'all', NONFATAL => 'redefine';
 use parent qw{Exporter};
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 
 =head1 NAME
 
@@ -12,16 +12,29 @@ EntityModel::Log - simple logging support for L<EntityModel>
 
 =head1 VERSION
 
-version 0.003
+version 0.004
 
 =head1 SYNOPSIS
 
  use EntityModel::Log ':all';
+ # Log everything down to level 0 (debug)
+ EntityModel::Log->instance->min_level(0);
+
+ # STDERR by default, or Test::More::note if you have it loaded
  logDebug("Test something");
  logInfo("Object [%s] found", $obj->name);
  logError("Fatal problem");
  logInfo(sub { my $str = heavy_operation(); return 'Failed: %s', $str });
 
+ logInfo("Stack trace - note that it must have at least one parameter (%s): %S", 'like this');
+ logInfo("No stack trace without parameters despite %S");
+
+ my $log = EntityModel::Log->instance;
+ $log->debug("OO-style debug");
+ $log->info("OO-style info");
+ $log->warning("OO-style warning");
+ $log->error("OO-style error");
+
 =head1 DESCRIPTION
 
 Yet another logging class. Provides a procedural and OO interface as usual - intended for use
@@ -37,14 +50,24 @@ use Time::HiRes qw{time};
 use POSIX qw{strftime};
 use Exporter;
 use List::Util qw{min max};
+use Scalar::Util qw{blessed};
 use IO::Handle;
+use File::Basename ();
 use Data::Dump ();
+use Data::Dump::Filtered ();
 
-our %EXPORT_TAGS = ( 'all' => [qw/&logDebug &logInfo &logWarning &logError &logStack/] );
+our %EXPORT_TAGS = ( 'all' => [qw/&logDebug &logInfo &logWarning &logError/] );
 our @EXPORT_OK = ( @{$EXPORT_TAGS{'all'}} );
 
 # Internal singleton instance 
 my $instance;
+
+=head2 instance
+
+Returns a handle to the main instance of L<EntityModel::Log>.
+
+=cut
+
 sub instance { my $class = shift; $instance ||= $class->new }
 
 =head1 PROCEDURAL METHODS
@@ -55,19 +78,26 @@ my @LogType = (
 	'Debug',
 	'Info',
 	'Warning',
-	'ERROR'
+	'Error',
+	'Fatal',
 );
 
-sub logBase { __PACKAGE__->instance->raise(@_); }
+=head2 _raise_error_on_global_instance
 
-=head2 logDebug
-
-Raise a debug message, but only if the debug flag is set. Expect a high volume of these during normal operation
-so live server has this switched off.
+Raise the given (code, message, ...) log event on the L<EntityModel::Log> global instance.
 
 =cut
 
-sub logDebug { unshift @_, 0; goto &logBase; }
+sub _raise_error_on_global_instance { __PACKAGE__->instance->raise(@_); }
+
+=head2 logDebug
+
+Raise a debug message. Expect a high volume of these during normal operation
+so a production server would typically have these disabled.
+
+=cut
+
+sub logDebug { unshift @_, 0; goto &_raise_error_on_global_instance; }
 
 =head2 logInfo
 
@@ -76,16 +106,17 @@ reasons - indicates normal operations rather than an error condition.
 
 =cut
 
-sub logInfo { unshift @_, 1; goto &logBase; }
+sub logInfo { unshift @_, 1; goto &_raise_error_on_global_instance; }
 
 =head2 logWarning
 
-Raise a warning message, for things like 'article does not exist', expect a few of these in regular operations
-but they aren't serious enough to be potential bugs or system problems.
+Raise a warning message, for things like 'requested delete for object that does not exist'.
+You might expect a few of these in regular operations due to concurrent access and timing issues,
+so they may not necessarily indicate real system problems.
 
 =cut
 
-sub logWarning { unshift @_, 2; goto &logBase; }
+sub logWarning { unshift @_, 2; goto &_raise_error_on_global_instance; }
 
 =head2 logError
 
@@ -93,7 +124,7 @@ Raise an error - this is likely to be a genuine system problem.
 
 =cut
 
-sub logError { unshift @_, 3; goto &logBase; }
+sub logError { unshift @_, 3; goto &_raise_error_on_global_instance; }
 
 =head2 logStack
 
@@ -102,50 +133,71 @@ Raise an error with stack - this is likely to be a genuine system problem.
 =cut
 
 sub logStack {
-	my $txt = __PACKAGE__->instance->parseMessage(@_);
+	my $txt = __PACKAGE__->instance->parse_message(@_);
 
 	$txt .= join("\n", map {
 		sprintf("%s:%s %s", $_->{filename}, $_->{line}, $_->{subroutine})
-	} stackTrace());
-	logBase(3, $txt);
+	} _stack_trace());
+	_raise_error_on_global_instance(3, $txt);
 }
 
-=head2 stackTrace
+=head2 _stack_trace
 
 Get a stack trace, as an array of hashref entries, skipping the top two levels.
 
 =cut
 
-sub stackTrace {
+sub _stack_trace {
+	my $skip = shift || 0;
+	my $dump = shift || 0;
 	my $idx = 1;
 	my @trace;
 	my $basePath = '';
-	while($idx < 99 && (my @stack = caller($idx))) {
-		my %info;
-                foreach (qw/package filename line subroutine hasargs wantarray evaltext is_require hints bitmask hinthash/) {
-			$info{$_} = (shift(@stack) // '');
+	{
+		package DB;
+		while($idx < 99 && (my @stack = caller($idx))) {
+			++$idx;
+			next if $skip-- > 0;
+
+			my %info;
+			@info{qw/package filename line subroutine hasargs wantarray evaltext is_require hints bitmask hinthash/} = map $_ // '', @stack;
+			$info{args} = [ @DB::args ];
+			push @trace, \%info;
 		}
-		if(0) { # could include source context using something like $info{filename} ~~ m{^$basePath/(.*)$} || $info{filename} ~~ m{^/perl-module-path/(.*)$}) {
-			my $file = $1;
-			if(-r $info{filename}) {
-				my $start = max(0, ($info{line} // 0) - 5);
-				$info{code} = '';
-				open my $fh, '<', $info{filename} or die $!;
+	}
+
+	foreach my $info (@trace) {
+		$info->{file} = File::Basename::basename($info->{filename});
+		$info->{code} = '';
+		if($dump) { # could include source context using something like $info{filename} ~~ m{^$basePath/(.*)$} || $info{filename} ~~ m{^/perl-module-path/(.*)$}) {
+			# I'm hoping this entire function can be replaced by a module from somewhere
+			if(-r $info->{filename}) {
+				# Start from five lines before the required line, but clamp to zero
+				my $start = max(0, ($info->{line} // 0) - 5);
+
+				# Probably not a safe thing to do, but most modules seem to be ascii or utf8
+				open my $fh, '<:encoding(utf8)', $info->{filename} or die $! . ' when reading ' . $info->{filename} . ' which we expected to have loaded already';
+
+				# 
 				if($start) {
-					<$fh> for 0..$start;
+					<$fh> for 1..$start;
 				}
-				$info{code} .= sprintf("%5d %s", $fh->input_line_number + 1, scalar(<$fh>)) for 0..10;
+				my $line = $start;
+				$info->{code} .= sprintf("%5d %s", $line++, <$fh> // last) for 0..10;
 				close $fh;
 			}
-			$info{filename} = $file;
 		}
-		push @trace, \%info;
-		++$idx;
 	}
-	return @trace;
+	return reverse @trace;
 }
 
-sub levelFromString {
+=head2 _level_from_string
+
+Returns the level matching the given string.
+
+=cut
+
+sub _level_from_string {
 	my $str = lc(shift);
 	my $idx = 0;
 	foreach (@LogType) {
@@ -155,7 +207,13 @@ sub levelFromString {
 	die "Bad log level [$str]";
 }
 
-sub timestamp {
+=head2 _timestamp
+
+Generate a string in ISO8601-ish format representing the time of this log event.
+
+=cut
+
+sub _timestamp {
 	my $now = Time::HiRes::time;
 	return strftime("%Y-%m-%d %H:%M:%S", gmtime($now)) . sprintf(".%03d", int($now * 1000.0) % 1000.0);
 }
@@ -170,7 +228,39 @@ Constructor - currently doesn't do much.
 
 =cut
 
-sub new { bless { handle => \*STDERR, isOpen => 1, pid => $$ }, shift }
+sub new { bless { handle => undef, is_open => 1, pid => $$ }, shift }
+
+=head2 debug
+
+Display a debug message.
+
+=cut
+
+sub debug { shift->raise(0, @_) }
+
+=head2 info
+
+Display an info message.
+
+=cut
+
+sub info { shift->raise(1, @_) }
+
+=head2 warning
+
+Display a warning message.
+
+=cut
+
+sub warning { shift->raise(2, @_) }
+
+=head2 error
+
+Display an error message.
+
+=cut
+
+sub error { shift->raise(3, @_) }
 
 =head2 path
 
@@ -181,7 +271,7 @@ Accessor for path setting, if given a new path will close existing file and dire
 sub path {
 	my $self = shift;
 	if(@_) {
-		$self->close if $self->isOpen;
+		$self->close if $self->is_open;
 		$self->{path} = shift;
 		$self->open;
 		return $self;
@@ -189,24 +279,6 @@ sub path {
 	return $self->{path};
 }
 
-=head2 handle
-
-Direct accessor for the file handle.
-
-=cut
-
-sub handle {
-	my $self = shift;
-	if(@_) {
-		$self->close if $self->isOpen;
-		$self->{handle} = shift;
-		$self->isOpen(1);
-		$self->pid($$);
-		return $self;
-	}
-	return $self->{handle} || $self->reopen;
-}
-
 =head2 pid
 
 Current PID, used for fork tracking.
@@ -222,19 +294,19 @@ sub pid {
 	return $self->{pid};
 }
 
-=head2 isOpen
+=head2 is_open
 
 Returns true if our log file is already open.
 
 =cut
 
-sub isOpen {
+sub is_open {
 	my $self = shift;
 	if(@_) {
-		$self->{isOpen} = shift;
+		$self->{is_open} = shift;
 		return $self;
 	}
-	return $self->{isOpen};
+	return $self->{is_open};
 }
 
 =head2 disabled
@@ -259,14 +331,13 @@ Close the log file if it's currently open.
 
 sub close : method {
 	my $self = shift;
-	return $self unless $self->isOpen;
-	if(my $h = $self->handle) {
-		close $h or die "Failed to close log file: $!\n";
+	return $self unless $self->is_open;
+
+	if(my $h = delete $self->{handle}) {
+		# Mark as closed so we try to open a new handle even if we break this one off here
+		$self->is_open(0);
+		$h->close or die "Failed to close log file: $!\n";
 	}
-
-	delete $self->{handle};
-
-	$self->isOpen(0);
 	return $self;
 }
 
@@ -278,7 +349,7 @@ Close any active handle if we've forked. This method just does the closing, not
 
 sub close_after_fork {
 	my $self = shift;
-	return unless $self->isOpen;
+	return unless $self->is_open;
 
 # Don't close STDOUT/STDERR. Bit of a hack really, we should perhaps just close when we were given a path?
 	return if $self->handle ~~ [\*STDERR, \*STDOUT];
@@ -294,7 +365,7 @@ Open the logfile.
 
 sub open : method {
 	my $self = shift;
-	return $self if $self->isOpen;
+	return $self if $self->is_open;
 	open my $fh, '>>', $self->path or die $! . " for " . $self->path;
 	binmode $fh, ':encoding(utf-8)';
 	$fh->autoflush(1);
@@ -310,12 +381,12 @@ Helper method to close and reopen logfile.
 
 sub reopen {
 	my $self = shift;
-	$self->close if $self->isOpen;
+	$self->close if $self->is_open;
 	$self->open;
 	return $self;
 }
 
-=head2 parseMessage
+=head2 parse_message
 
 Generate appropriate text based on whatever we get passed.
 
@@ -327,8 +398,6 @@ Item parsing handles the following types:
 
 =item * Single string is passed through unchanged
 
-=item * Any coderef is expanded in place (recursively - a coderef can return other coderefs)
-
 =item * Arrayref or hashref is expanded via L<Data::Dump>
 
 =item * Other references are stringified
@@ -337,12 +406,18 @@ Item parsing handles the following types:
 
 =back
 
+In addition, if the first parameter is a coderef then it is expanded in place (recursively - a coderef can return another coderef). Note that this only happens for the *first* parameter at each
+level of recursion.
+
 =cut
 
-sub parseMessage {
+sub parse_message {
 	my $self = shift;
+#	warn $_[0];
 	return '' unless @_;
 
+	unshift @_, $_[0]->() while ref($_[0]) ~~ /^CODE/;
+
 # Decompose parameters into strings
 	my @data;
 	ITEM:
@@ -368,6 +443,23 @@ sub parseMessage {
 	my $fmt = shift(@data) // '';
 	return $fmt unless @data;
 
+	# Special-case the stack trace feature. A bit too special really :(
+	$fmt =~ s/%S/join "\n", '', map {
+		my $txt = sprintf ' => %-32.32s %s(%s)',
+			$_->{package} . ':' . $_->{line},
+			($_->{subroutine} =~ m{ ( [^:]+$ ) }x),
+			  ($_->{package} ~~ 'EntityModel::Log')
+			? ('')
+			: (join ', ', map Data::Dump::Filtered::dump_filtered($_, sub {
+				my ($ctx, $obj) = @_;
+				return undef unless $ctx->is_blessed;
+				return { dump => "$obj" };
+			}), @{ $_->{args} });
+		$txt =~ s{%}{%%}g;
+		$txt;
+	} _stack_trace(0, 1)/e;
+	die "Format undef" unless defined $fmt;
+	die "Undefined entry in data, others are " . join ', ', map { defined($_) } @data if grep { !defined($_) } @data;
 	return sprintf($fmt, @data);
 }
 
@@ -411,26 +503,69 @@ sub raise {
 	$minLevel //= 1;
 	return $self if $minLevel > $level;
 
-	my $txt = $self->parseMessage(@_);
+	my $txt = $self->parse_message(@_);
 
 # Explicitly get time from Time::HiRes for ms accuracy
-	my $ts = timestamp();
+	my $ts = _timestamp();
 
 	my $type = sprintf("%-8.8s", $LogType[$level]);
-	$self->close_after_fork unless $$ ~~ $self->pid;
-	$self->open unless $self->isOpen;
-	$self->handle->print("$ts $type $file:$line $txt\n");
+	$self->output("$ts $type $file:$line $txt");
 	return $self;
 }
 
-=head2 debug
-
-Log a message at debug level.
+=head2 output
 
 =cut
 
-sub debug {
+sub output {
 	my $self = shift;
+	my $msg = shift;
+	if(my $handle = $self->get_handle) {
+		$handle->print($msg . "\n");
+		return $self;
+	}
+
+	Test::More::note($msg);
+	return $self;
+}
+
+=head2 get_handle
+
+Returns a handle if we have one, and 0 if we should fall back to L<Test::More>::note.
+
+=cut
+
+sub get_handle {
+	my $self = shift;
+	# Fall back to Test::More if available, unless we already have a handle
+	if(!$self->{handle}) {
+		return 0 if $ENV{HARNESS_ACTIVE};
+		# Exists, but undef, means STDERR fallback
+		return \*STDERR if exists $self->{handle};
+	}
+
+	$self->close_after_fork unless $$ ~~ $self->pid;
+
+	$self->open unless $self->is_open;
+	return $self->handle;
+}
+
+=head2 handle
+
+Direct(-ish) accessor for the file handle.
+
+=cut
+
+sub handle {
+	my $self = shift;
+	if(@_) {
+		$self->close if $self->is_open;
+		$self->{handle} = shift;
+		$self->is_open(1);
+		$self->pid($$);
+		return $self;
+	}
+	return $self->{handle} || $self->reopen;
 }
 
 END { $instance->close if $instance; }
